/***************************************************************************
 * This file is part of the BF527 doorbell application.                    *
 * Doorbell is a VOIP-based outdoor doorbell that can be run on SIP-based  *
 * telephone systems (PBX).                                                *
 *                                                                         *
 * Copyright (C) 2009,2010 by BerLinux Solutions GmbH                      *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at your  *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.           *
 ***************************************************************************/

#include <cstdio>
#include <disko.h>
#include <pthread.h>
#include "doorbell.h"
#include "sipbackend_minisip.h"
#include "sipbackend_linphone.h"
#include "sunphase.h"
#include "themeswitchthread.h"

extern void playIntro();

#ifdef __MTRACE__
#include <mcheck.h>
#endif

static bool signin=true;

#if 0
static void cleanUp(void *ptr) {
	if(ptr) {
		DEBUGMSG("doorbell", "cleaning dialog %p", ((CleanUpPtr*)ptr)->dialog);
		switch(((CleanUpPtr*)ptr)->type) {
			case UNKNOWN_DIALOG:
				break;
			case PIN_DIALOG:
				delete (PinDialog *)((CleanUpPtr*)ptr)->dialog;
				break;
			case ERROR_DIALOG:
				delete (ErrorDialog *)((CleanUpPtr*)ptr)->dialog;
				break;
			case ABSENCE_DIALOG:
				delete (AbsenceDialog *)((CleanUpPtr*)ptr)->dialog;
				break;
			case RINGING_DIALOG:
				delete (RingingDialog *)((CleanUpPtr*)ptr)->dialog;
				break;
			case VOLUME_DIALOG:
				delete (VolumeDialog *)((CleanUpPtr*)ptr)->dialog;
				break;
		}

		delete ((CleanUpPtr*)ptr);
	}
}
#endif


DoorBell::DoorBell() :
		mainMenu(NULL),
		currentDialog(NULL),
		pinDialog(NULL),
		errorDialog(NULL),
		absenceDialog(NULL),
		ringingDialog(NULL),
		volumeDialog(NULL),
		sipController(NULL) {

	DataSource datasource(config.getConfigDBDBMS(),
						  config.getConfigDBDatabase(),
						  config.getConfigDBAddress(),
						  config.getConfigDBPort(),
						  config.getConfigDBUser(),
						  config.getConfigDBPassword());

	MMSPluginService service(&datasource);
	MMSPROPERTYDATA_MAP props;
	service.getSystemProperties(props);

	MMSPROPERTYDATA_MAP::iterator prop = props.find("LodgerList");

	if(prop == props.end()) {
		fprintf(stderr, "No LodgerList definded!\n");
		exit(1);
	}

	UserFileReader r(prop->second->getValue());
	r.read();
	vector<User *> u = r.getUsers();
	for(size_t i = 0; i < u.size(); i++) {
		users[u[i]->name] = u[i];
	}

	for(MMSPROPERTYDATA_MAP::iterator it = props.begin(); it != props.end();it++) {
		delete it->second;
	}

	this->mainMenu = new MainMenu(&users);
	this->mainMenu->lodgerMenu->onReturn->connect(
			sigc::mem_fun(this, &DoorBell::onLodgerSelect));
	this->mainMenu->lodgerMenu->onSelectItem->connect(
			sigc::mem_fun(this, &DoorBell::onSelectItemMainMenu));
	this->mainMenu->pinButton->onReturn->connect(
			sigc::mem_fun(this, &DoorBell::onPinSelect));

	this->currentDialog = this->mainMenu;

	showMenuTimer = new MMSTimer(false);
	showMenuTimer->timeOut.connect(sigc::mem_fun(this, &DoorBell::onTimeOut));
	showMenuTimer->start(10000);

	if(!signin) {
		this->mainMenu->show();
	}

	try{
#if defined(__USE_MINISIP__)
		this->sipController = new SipbackendMinisip();
#elif defined(__USE_LINPHONE__)
		this->sipController = new SipbackendLinphone();
#endif
		if(this->sipController) {
			this->sipController->onEstablished.connect(sigc::mem_fun(this, &DoorBell::onEstablished));
			this->sipController->onDisconnect.connect(sigc::mem_fun(this, &DoorBell::onDisconnect));
			this->sipController->onError.connect(sigc::mem_fun(this, &DoorBell::onError));
			this->sipController->onRegister.connect(sigc::mem_fun(this, &DoorBell::onRegister));
			if(signin) {
				this->sipController->initialize();
			}
		}

#ifdef __EPCBF527__
		this->gpioThread = new GPIOThread(this->sipController);
		this->gpioThread->start();
#endif
	} catch(MMSError *e) {
		string msg = e->getMessage();
		sendErrorMail(msg);
		showErrorDialog(msg);
		delete e;
	}
}



DoorBell::~DoorBell() {
#ifdef __EPCBF527__
	if(this->gpioThread) {
		delete this->gpioThread;
	}
#endif
	if(this->sipController) {
		delete this->sipController;
	}
	delete showMenuTimer;
	delete mainMenu;
}



void DoorBell::start() {
	pause();
}



void DoorBell::onSelectItemMainMenu(MMSWidget *wid) {
	showMenuTimer->restart();
}



void DoorBell::onLodgerSelect(MMSWidget *wid) {
	MMSLabelWidget *lodgerName;
	User *u;
	MMSWidget *lodger;

	if(!(lodger = this->mainMenu->lodgerMenu->getSelectedItem())) {
		return;
	}

	if(!(lodgerName = dynamic_cast<MMSLabelWidget *>(lodger->searchForWidget("nameLabel")))) {
		return;
	}

	if(!(u = users.at(lodgerName->getText()))) {
		return;
	}

	if(this->sipController) {
		if(!u->awayMsg.empty()) {
			// the lodger is abscent
			if(!this->absenceDialog) {
				this->absenceDialog = new AbsenceDialog(u, sipController);
			} else {
				this->absenceDialog->setUser(u);
			}
			this->currentDialog = this->absenceDialog;
		} else {
			// the lodger is present
			if(!this->ringingDialog) {
				this->ringingDialog = new RingingDialog(u, sipController);
			} else {
				this->ringingDialog->setUser(u);
			}
			this->currentDialog = this->ringingDialog;
		}

		this->currentDialog->onExit.connect(
					sigc::mem_fun(this, &DoorBell::onHideDialog));
	}
	
	this->mainMenu->hide();
	this->currentDialog->show();
}



void DoorBell::onPinSelect(MMSWidget *wid) {
	this->showMenuTimer->restart();

	this->pin = "";
	if(!this->pinDialog) {
		this->pinDialog = new PinDialog();
		this->pinDialog->onExit.connect(
				sigc::mem_fun(this, &DoorBell::onHideDialog));

		this->pinDialog->pad1->onReturn->connect(
				sigc::mem_fun(this, &DoorBell::onReturnPinDialog));
		this->pinDialog->pad2->onReturn->connect(
				sigc::mem_fun(this, &DoorBell::onReturnPinDialog));
		this->pinDialog->pad3->onReturn->connect(
				sigc::mem_fun(this, &DoorBell::onReturnPinDialog));
		this->pinDialog->pad4->onReturn->connect(
				sigc::mem_fun(this, &DoorBell::onReturnPinDialog));
		this->pinDialog->pad5->onReturn->connect(
				sigc::mem_fun(this, &DoorBell::onReturnPinDialog));
		this->pinDialog->pad6->onReturn->connect(
				sigc::mem_fun(this, &DoorBell::onReturnPinDialog));
	}

	this->currentDialog = this->pinDialog;

	this->mainMenu->hide();
	this->pinDialog->show();

	if(this->sipController) {
		this->sipController->hangup();
	}
}



void DoorBell::showErrorDialog(std::string errorMsg,
		bool isCritical,
		std::string timerMsg,
		int timerLength) {

	DEBUGMSG("doorbell", "isCritical = %s", (isCritical ? "true" : "false"));
	DEBUGMSG("doorbell", "errorMsg = %s - this->errorMsg = %s", errorMsg.c_str(), this->errorMsg.c_str());
	DEBUGMSG("doorbell", "this->currentDialog = %p - this->errorDialog = %p", this->currentDialog, this->errorDialog);
	
	if(errorMsg == this->errorMsg
			&& this->currentDialog == this->errorDialog
			&& timerMsg == this->timerMsg
			&& timerLength == this->timerLength) {
		DEBUGMSG("doorbell", "skipping error dialog (already shown)");
		this->errorDialog->showTimedMsg();
		return;
	}
	
	DEBUGMSG("doorbell", "showing error dialog");

	if(this->currentDialog && (this->currentDialog != this->mainMenu)) {
		onHideDialog(this->currentDialog, false);
	}

	this->errorMsg    = errorMsg;
	this->timerMsg    = timerMsg;
	this->timerLength = timerLength;


	if(isCritical) {
		this->showMenuTimer->stop();
	} else {
		this->showMenuTimer->restart();
	}

	if(!this->errorDialog) {
		this->errorDialog   = new ErrorDialog(errorMsg, timerMsg, timerLength);
		this->errorDialog->onExit.connect(
					sigc::mem_fun(this, &DoorBell::onHideDialog));
	} else {
		this->errorDialog->update(errorMsg, timerMsg, timerLength);
	}
	this->currentDialog = this->errorDialog;

	if(isCritical) {
		this->errorDialog->cancelBtn->setClickable(false);
		this->errorDialog->cancelBtn->setVisible(false, true);
	} else {
		this->errorDialog->cancelBtn->setClickable(true);
		this->errorDialog->cancelBtn->setVisible(true, true);
	}
	this->errorDialog->show();
	this->errorDialog->showTimedMsg();

}

void DoorBell::onHideDialog(Dialog *dialog, bool showMainWindow) {

	DEBUGMSG("doorbell", "hiding dialog %p", dialog);

	if((dialog == NULL) || (dialog == this->mainMenu)) {
		return;
	}
//	CleanUpPtr *ptr = new CleanUpPtr;
//	ptr->type = UNKNOWN_DIALOG;
//	ptr->dialog = NULL;

	string errMsg = "";

	if(dialog == this->pinDialog) {
		DEBUGMSG("doorbell", "hiding pin dialog");
		if(this->pin.length() == 4) {
			bool found = false;
			std::map<string, User*>::iterator it;
			for(it = users.begin(); it != users.end(); ++it) {
				if(it->second->code == pin) {
					printf("OPENING DOOR FOR: %s\n", it->second->name.c_str());
					found = true;
					if(this->sipController) {
						this->sipController->setVolume(100);
					}
					if(!system("scripts/opendoor.sh")) {
						break;
					}
					break;
				}
			}

			if(!found) {
				this->pin = "";
				errMsg = "Sorry, wrong code.";
			}
		}

//		ptr->type       = PIN_DIALOG;
//		ptr->dialog     = this->pinDialog;
//		this->pinDialog = NULL;
	}
/*	else if(dialog == this->errorDialog) {
		ptr->type         = ERROR_DIALOG;
		ptr->dialog       = this->errorDialog;
		this->errorDialog = NULL;
	}
	else if(dialog == this->absenceDialog) {
		ptr->type           = ABSENCE_DIALOG;
		ptr->dialog         = this->absenceDialog;
		this->absenceDialog = NULL;
	}
	else if(dialog == this->ringingDialog) {
		ptr->type           = RINGING_DIALOG;
		ptr->dialog         = this->ringingDialog;
		this->ringingDialog = NULL;
	}
	else if(dialog == this->volumeDialog) {
		ptr->type          = VOLUME_DIALOG;
		ptr->dialog        = this->volumeDialog;
		this->volumeDialog = NULL;
	}
	addGarbageHandler(cleanUp,ptr);
*/
	this->showMenuTimer->restart();
	this->currentDialog = this->mainMenu;
	
	if(!errMsg.empty()) {
		showErrorDialog(errMsg, false);
	} else {
		if(showMainWindow) {
			this->mainMenu->show();
		}
	}
}



void DoorBell::onError(std::string errorMsg, bool isCritical,
		std::string timerMsg, int length) {
	if(!errorMsg.empty()) {
		sendErrorMail(errorMsg);
		showErrorDialog(errorMsg, isCritical, timerMsg, length);
	}

	if(this->sipController) {
		DEBUGMSG("doorbell", "sending hangup");
		this->sipController->hangup();
	}
}



void DoorBell::onRegister(RegisterStatus status) {
	switch(status) {
		case TRYING:
			showErrorDialog("Register...\nPlease wait!", true, "", 0);
			break;
		case FAILED:
			showErrorDialog("Registration failed", true, "Retry in %d s", 3);
			break;
		case OK:
			if(this->currentDialog == this->errorDialog) {
				this->errorDialog->hide();
			}
			break;
	}
}



void DoorBell::onReturnPinDialog(MMSWidget *widget) {
	this->showMenuTimer->restart();

	if(widget == this->pinDialog->pad1) {
		this->pin += "1";
	} else if(widget == this->pinDialog->pad2) {
		this->pin += "2";
	} else if(widget == this->pinDialog->pad3) {
		this->pin += "3";
	} else if(widget == this->pinDialog->pad4) {
		this->pin += "4";
	} else if(widget == this->pinDialog->pad5) {
		this->pin += "5";
	} else if(widget == this->pinDialog->pad6) {
		this->pin += "6";
	}

	if(this->pin.length() == 4) {
		this->pinDialog->hide();
	}
}



void DoorBell::onTimeOut() {
	if(this->currentDialog == this->mainMenu) {
		this->mainMenu->rewind();
	} else if(this->currentDialog == this->errorDialog) {
		DEBUGMSG("doorbell", "timeout for errorDialog");
		onHideDialog(this->errorDialog);
	} else if(this->currentDialog == this->pinDialog) {
		DEBUGMSG("doorbell", "timeout for pinDialog");
		onHideDialog(this->pinDialog);
	}
}



void DoorBell::onEstablished(void *) {
	if(this->currentDialog == this->ringingDialog
			|| this->currentDialog == this->absenceDialog) {
		onHideDialog(this->currentDialog, false);
	}
	
	if(!this->volumeDialog) {
		this->volumeDialog  = new VolumeDialog(sipController);
		this->volumeDialog->onExit.connect(
				sigc::mem_fun(this, &DoorBell::onHideDialog));
	}
	this->currentDialog = this->volumeDialog;
	this->currentDialog->show();
}



void DoorBell::onDisconnect(void *) {
	this->errorMsg = "";
	onHideDialog(this->currentDialog);
}

bool DoorBell::inCall() {
	if(sipController) {
		return sipController->inCall();
	} else {
		return false;
	}
}

void DoorBell::sendErrorMail(std::string &errMsg) {
#ifdef __HAVE_VMIME__
	DataSource datasource(config.getConfigDBDBMS(),
						  config.getConfigDBDatabase(),
						  config.getConfigDBAddress(),
						  config.getConfigDBPort(),
						  config.getConfigDBUser(),
						  config.getConfigDBPassword());

	MMSPluginService service(&datasource);
	MMSPROPERTYDATA_MAP props;
	service.getSystemProperties(props);

	MMSPROPERTYDATA_MAP::iterator mailserver = props.find("Mailserver");
	MMSPROPERTYDATA_MAP::iterator user = props.find("MailserverUsername");
	MMSPROPERTYDATA_MAP::iterator pwd = props.find("MailserverPwd");
	MMSPROPERTYDATA_MAP::iterator to = props.find("AdminMailTo");
	MMSPROPERTYDATA_MAP::iterator from = props.find("AdminMailFrom");

	MMSMail mail("An error Occured",from->second->getValue(),mailserver->second->getValue());
	mail.setAuthData(user->second->getValue(), pwd->second->getValue());
	mail.addRecipient(to->second->getValue());
	mail.setMailBody("The doorbell experienced an error: " + errorMsg);
	try {
		mail.send();
	} catch(MMSError *err) {
		DEBUGMSG("doorbell", "Error Mail sending failed: %s", err->getMessage().c_str());
	}
#endif
}

static DoorBell     *app = NULL;
static bool night =  false;

static void themeSwitch(int sig) {
	printf("got sigusr1\n");
	signal (SIGALRM, themeSwitch);
	return;
}

static void myExit() {
	if(app) {
		delete app;
	}

#ifdef __MTRACE__
	muntrace();
#endif
}

void determineTheme(string &theme) {
	MMSConfigData config;
	DataSource datasource(config.getConfigDBDBMS(),
						  config.getConfigDBDatabase(),
						  config.getConfigDBAddress(),
						  config.getConfigDBPort(),
						  config.getConfigDBUser(),
						  config.getConfigDBPassword());

	MMSPluginService service(&datasource);
	MMSPROPERTYDATA_MAP props;
	service.getSystemProperties(props);

	MMSPROPERTYDATA_MAP::iterator longi = props.find("Longitude");
	MMSPROPERTYDATA_MAP::iterator lati = props.find("Latitude");

	double longitude = strtod(longi->second->getValue().c_str(), NULL);
	double lattitude = strtod(lati->second->getValue().c_str(), NULL);

	time_t rise, set, now;
	now=time(NULL);
	calcSunPhase(rise,set,now,longitude,lattitude);
	if((now >= set)||(now <= rise)) {
		theme="night";
	} else if ((now > rise) && (now < set)) {
		theme="day";
	}
//	printf("sunrise: %d\nnow:     %d\nsunset:  %d\n", rise,now,set);

}

int main(int argc, char *argv[]) {
	string diskorc;
	bool intro=true;
	//time_t rise,set,now;
	bool override=false;
	string theme;
	bool testswitch = false;
	bool calltest = false;

#ifdef __MTRACE__
	mtrace();
#endif

	diskorc="./etc/diskorc_day.xml";

	for(int i = 1; i < argc; i++) {
		if(strcmp(argv[i], "-x11") == 0) {
			diskorc="./etc/diskorc_x11.xml";
		} else if(strcmp(argv[i], "-night") == 0) {
			night=true;
			override=true;
			diskorc="./etc/diskorc_night.xml";
		} else if(strcmp(argv[i], "-day") == 0) {
			night=false;
			override=true;
			diskorc="./etc/diskorc_day.xml";
		} else if(strcmp(argv[i],"-nointro")==0) {
			intro = false;
		} else if(strcmp(argv[i],"-nosign")==0) {
			signin = false;
		} else if(strcmp(argv[i],"-calltest")==0) {
			calltest = true;
		} else if(strcmp(argv[i],"-testswitch")==0) {
			testswitch = true;
		}
	}

	if(calltest) {
#if defined(__USE_MINISIP__)
		Sipbackend *sip = new SipbackendMinisip();
#elif defined(__USE_LINPHONE__)
		Sipbackend *sip = new SipbackendLinphone();
#endif
		if(!sip) {
			cerr << "Error initializing SIP!" << endl;
			return -1;
		}
		sip->initialize();
		sleep(5);
		for(int i = 0; i < 5; ++i) {
			sip->call("600", "");
			sleep(5);
			sip->hangup();
			sleep(2);
		}
		return 0;
	}

	if(!mmsInit(MMSINIT_EVENTS|MMSINIT_GRAPHICS|MMSINIT_INPUTS|MMSINIT_THEMEMANAGER|MMSINIT_WINDOWMANAGER|MMSINIT_WINDOWS, argc, argv, diskorc,
				 "Doorbell", "BerLinux")) {

		fprintf(stderr, "Disko intialization failed!\n");
		exit(1);
	}

	if(!override) {
		determineTheme(theme);
		MMSThemeManager tm;
		tm.setTheme(theme,false);
	}

	if(intro) {
		playIntro();
	}

	app = new DoorBell();

	if(!app) {
		fprintf(stderr, "Error starting doorbell!\n");
		exit(1);
	}

	atexit(myExit);
	sleep(5);
	if(override) {
		if(night == false) {
			theme = "day";
		} else {
			theme = "night";
		}
	}
	ThemeSwitchThread *tsw = new ThemeSwitchThread(theme);
	tsw->start();

	if(testswitch) {
		while(1) {
			sleep(10);
			MMSThemeManager tm;
			if(!night) {
				theme = "night";
				night=true;
			} else {
				theme = "day";
				night=false;
			}
			tm.setTheme(theme,true);
		}
	}

	app->start();

	return 0;
}

