/***************************************************************************
 * This file is part of the BF527 doorbell application.                    *
 * Doorbell is a VOIP-based outdoor doorbell that can be run on SIP-based  *
 * telephone systems (PBX).                                                *
 *                                                                         *
 * Copyright (C) 2009,2010 by BerLinux Solutions GmbH                      *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at your  *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.           *
 ***************************************************************************/

#include "sipbackend.h"

#include <libxml/parser.h>

#ifndef __HAVE_MIXER__
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#endif

#include <mmstools/tools.h>

Sipbackend::Sipbackend() :
	stunserverIP(""),
	nameserver(""),
	interface(""),
	audioInDevice(""),
	audioOutDevice(""),
	ringtone(""),
	busytone(""),
	calltone(""),
	volume(100),
	stunserverPort(0),
	account(),
	accountRegistered(false) {

};

bool Sipbackend::initialize() {
	if(!readConfig()) {
		onError.emit("Configuration error!", true, "", 0);
		return false;
	}
	onRegister.emit(TRYING);
	return true;
}

/*
 * Read configuration file.
 *
 * @note	The configuration file is named config.xml and is
 * located in the plugin folder.
 */
bool Sipbackend::readConfig() {
	this->configFile = "etc/config.xml";

	DEBUGMSG("SIPBACKEND", "Reading config file " + this->configFile);

	try {
		xmlDocPtr doc;

		LIBXML_TEST_VERSION

		doc = xmlReadFile(this->configFile.c_str(), NULL, 0);
		if(!doc)
			throw new MMSSipBackendError(1, "Could not parse file:" + this->configFile);

		/* Walk the tree: */
		xmlNodePtr rootNode = xmlDocGetRootElement(doc);
		if(xmlStrcmp(rootNode->name, (const xmlChar *) "doorbell")) {
			DEBUGMSG("SIPBACKEND", "invalid configuration file (" + this->configFile + ") - does not contain disko root node");
			throw new MMSSipBackendError(1, "Invalid configuration file:" + this->configFile);
		}

		xmlChar *parname, *parvalue;
		for(xmlNodePtr topNode = rootNode->xmlChildrenNode; topNode; topNode = topNode->next) {
			if(!xmlStrcmp(topNode->name, (const xmlChar*) "options")) {
				for (xmlNodePtr curNode = topNode->xmlChildrenNode; curNode; curNode = curNode->next) {
					if(xmlStrcmp(curNode->name, (const xmlChar*) "parameter"))	continue;
					if(!(parname  = xmlGetProp(curNode, (const xmlChar*)"name"))) continue;
					if(!(parvalue = xmlGetProp(curNode, (const xmlChar*)"value"))) continue;

					if(!xmlStrcmp(parname, (const xmlChar*)"ringtone")) this->ringtone = string((const char *)parvalue);
					else if(!xmlStrcmp(parname, (const xmlChar*)"busytone")) this->busytone = string((const char *)parvalue);
					else if(!xmlStrcmp(parname, (const xmlChar*)"calltone")) this->calltone = string((const char *)parvalue);
					else if(!xmlStrcmp(parname, (const xmlChar*)"audio_in_device")) this->audioInDevice = string((const char *)parvalue);
					else if(!xmlStrcmp(parname, (const xmlChar*)"audio_out_device")) this->audioOutDevice = string((const char *)parvalue);

					xmlFree(parname);
					xmlFree(parvalue);
				}
			}
			else if(!xmlStrcmp(topNode->name, (const xmlChar*) "connection")) {
				for (xmlNodePtr curNode = topNode->xmlChildrenNode; curNode; curNode = curNode->next) {
					if(xmlStrcmp(curNode->name, (const xmlChar*) "parameter"))	continue;
					if(!(parname  = xmlGetProp(curNode, (const xmlChar*)"name"))) continue;
					if(!(parvalue = xmlGetProp(curNode, (const xmlChar*)"value"))) continue;

					if(!xmlStrcmp(parname, (const xmlChar*)"interface")) this->interface = string((const char *)parvalue);
					else if(!xmlStrcmp(parname, (const xmlChar*)"stunserver_ip")) this->stunserverIP = string((const char *)parvalue);
					else if(!xmlStrcmp(parname, (const xmlChar*)"stunserver_port")) this->stunserverPort = strToInt((const char *)parvalue);
					else if(!xmlStrcmp(parname, (const xmlChar*)"nameserver")) this->nameserver = string((const char *)parvalue);

					xmlFree(parname);
					xmlFree(parvalue);
				}
			}
			else if(!xmlStrcmp(topNode->name, (const xmlChar*) "codecs")) {
				for (xmlNodePtr curNode = topNode->xmlChildrenNode; curNode; curNode = curNode->next) {
					if(xmlStrcmp(curNode->name, (const xmlChar*) "parameter")) continue;
					if(!(parname  = xmlGetProp(curNode, (const xmlChar*)"name"))) continue;
					if(!(parvalue = xmlGetProp(curNode, (const xmlChar*)"value"))) continue;

					if(!xmlStrcmp(parname, (const xmlChar*)"codec")) this->codecs.push_back(string((const char *)parvalue));

					xmlFree(parname);
					xmlFree(parvalue);
				}
			}
			else if(!xmlStrcmp(topNode->name, (const xmlChar*) "account") && account.user.empty()) {
				for (xmlNodePtr curNode = topNode->xmlChildrenNode; curNode; curNode = curNode->next) {
					if(xmlStrcmp(curNode->name, (const xmlChar*) "parameter"))	continue;
					if(!(parname  = xmlGetProp(curNode, (const xmlChar*)"name"))) continue;
					if(!(parvalue = xmlGetProp(curNode, (const xmlChar*)"value"))) continue;

					if(xmlStrEqual(parname, (const xmlChar*)"user"))           account.user      = string((const char*)parvalue);
					else if(xmlStrEqual(parname, (const xmlChar*)"passwd"))    account.passwd    = string((const char*)parvalue);
					else if(xmlStrEqual(parname, (const xmlChar*)"registrar")) account.registrar = string((const char*)parvalue);
					else if(xmlStrEqual(parname, (const xmlChar*) "realm"))    account.realm     = string((const char*)parvalue);

					xmlFree(parname);
					xmlFree(parvalue);
				}
			}
		}

		/*free the document */
		xmlFreeDoc(doc);
	}
	catch(MMSError *error)
	{
		DEBUGMSG("SIPBACKEND", "Exception caught: " + error->getMessage());
		throw new MMSSipBackendError(1, error->getMessage());
	}

	return true;
}

bool Sipbackend::setVolume(const unsigned int percent) {
#ifdef __HAVE_MIXER__
	if(!this->audioCtrl) {
		return false;
	}
#endif

	if(!isInitialized() || (percent > 100)) {
		return false;
	}

#ifdef __HAVE_MIXER__
	audioCtrl->setVolume(percent);
#else
	int fd = open("/dev/mixer", O_RDWR);
	if(fd < 0) {
		DEBUGMSG("SIPBACKEND", "error opening mixer device");
		return false;
	}

	/* setting to 100% */
	int value = (percent & 0xff);
	value |= (value << 8);
	if(ioctl(fd, SOUND_MIXER_WRITE_VOLUME, &value) == -1) {
		DEBUGMSG("SIPBACKEND", "error setting volume");
		close(fd);
		return false;
	}

	close(fd);
#endif

	return true;
}
