/***************************************************************************
 * This file is part of the BF527 doorbell application.                    *
 * Doorbell is a VOIP-based outdoor doorbell that can be run on SIP-based  *
 * telephone systems (PBX).                                                *
 *                                                                         *
 * Copyright (C) 2009,2010 by BerLinux Solutions GmbH                      *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at your  *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.           *
 ***************************************************************************/

#ifdef __USE_LINPHONE__

#define __USE_LINPHONE_CONFIG__

#include <mmstools/tools.h>
#include <mmsbase/mmsevent.h>

#include "sipbackend_linphone.h"

SipbackendLinphone::SipbackendLinphone() :
	Sipbackend(),
	linphoneInitialized(false),
	quit(false),
	inCallStatus(false),
	sipCore(NULL),
	sipProxy(NULL),
	sipAuth(NULL) {

}

SipbackendLinphone::~SipbackendLinphone() {
	shutdown();
}

static void _InviteCb(struct _LinphoneCore *lc, const char *from) {
	DEBUGMSG("SIPBACKEND", "invite received");
}

static void _GeneralStateCb(LinphoneCore *lc, LinphoneGeneralState *gstate) {
	cout << "gstate changed: " 
		 << iToStr(gstate->old_state) << "->" << iToStr(gstate->new_state)
		 << " - " << (gstate->message ? gstate->message : "") 
		 << endl;
	SipbackendLinphone *thiz = static_cast<SipbackendLinphone*>(linphone_core_get_user_data(lc));
	if(thiz) {
		thiz->handleState(gstate->new_state, gstate->message);
	}
}

static void _DisplayMessageCb(LinphoneCore *lc, const char *message) {
	DEBUGMSG("SIPBACKEND", "message: %s", message);
	cout << "message: " << message << endl;
}

static void _DisplayQuestionCb(LinphoneCore *lc, const char *question) {
	DEBUGMSG("SIPBACKEND", "question: %s", question);
	cout << "question: " << question << endl;
}

static void _DisplayStatusCb(LinphoneCore *lc, const char *status) {
	DEBUGMSG("SIPBACKEND", "status: %s", status);
	cout << "status: " << status << endl;

	SipbackendLinphone *thiz = static_cast<SipbackendLinphone*>(linphone_core_get_user_data(lc));
	if(thiz) {
		if(strcmp(status, "---Not found") == 0) {
			thiz->onError.emit("User not found!", false, "", 0);
		}
	}
}

static void _DisplayWarningCb(LinphoneCore *lc, const char *message) {
	DEBUGMSG("SIPBACKEND", "warning: %s", message);
	cout << "warning: " << message << endl;
}

static void _DisplayUrlCb(LinphoneCore *lc, const char *message, const char *url) {
	DEBUGMSG("SIPBACKEND", "URL: %s [%s]", message, url);
}

static void _AuthInfoRequestedCb(LinphoneCore *lc, const char *realm, const char *user) {
	DEBUGMSG("SIPBACKEND", "AuthInfo requested: %s [%s]", user, realm);
}

static void _ByeCb(LinphoneCore *lc, const char *from) {
	DEBUGMSG("SIPBACKEND", "bye received");
}

static void _ShowInterfaceCb(LinphoneCore *lc) {
}

static void _CallUpdatedCb(LinphoneCore *lc, _LinphoneCallLog *log) {
}

static void _DTMFReceived(LinphoneCore *lc, int dtmf) {
	if(dtmf == 11) {
		if(!system("scripts/blink.sh")) {
			DEBUGMSG("sipbackend", "error opening door");
		}
	}
}

#ifdef __ENABLE_LOG__
static void _LogCb(OrtpLogLevel lev, const char *fmt, va_list args) {
	vprintf(fmt, args);
	cout << endl;
}
#endif

bool SipbackendLinphone::initialize() {
	if(Sipbackend::initialize()) {
		((MMSThread*)this)->start();
		return true;
	}

	return false;
}

void SipbackendLinphone::initializeLinphone() {
	this->sipVTable.inv_recv = _InviteCb;
	this->sipVTable.general_state = _GeneralStateCb;
	this->sipVTable.display_message = _DisplayMessageCb;
	this->sipVTable.display_question = _DisplayQuestionCb;
	this->sipVTable.display_status = _DisplayStatusCb;
	this->sipVTable.display_warning = _DisplayWarningCb;
	this->sipVTable.display_url = _DisplayUrlCb;
	this->sipVTable.auth_info_requested = _AuthInfoRequestedCb;
	this->sipVTable.bye_recv = _ByeCb;
	this->sipVTable.show = _ShowInterfaceCb;
	this->sipVTable.call_log_updated = _CallUpdatedCb;
	this->sipVTable.dtmf_received = _DTMFReceived;

#ifdef __USE_LINPHONE_CONFIG__	
	this->sipCore = linphone_core_new(&this->sipVTable,	"./.linphonerc", NULL, this);
#else
	this->sipCore = linphone_core_new(&this->sipVTable,	"/dev/null", NULL, this);
#endif
	if(!this->sipCore) {
		throw MMSSipBackendError(0, "Could not initialize SIP");
	}
	

#ifdef __ENABLE_LOG__
	linphone_core_enable_logs_with_cb(_LogCb);
#else
	linphone_core_disable_logs();
#endif
	linphone_core_enable_video(this->sipCore, false, false);
	linphone_core_enable_video_preview(this->sipCore, false);

#ifndef __USE_LINPHONE_CONFIG__	

	if(!this->audioInDevice.empty()) {
		if(this->audioInDevice == "alsa:default") {
			this->audioInDevice = "ALSA: default device";
		} else if(this->audioInDevice == "/dev/dsp") {
			this->audioInDevice = "OSS: /dev/dsp";
		}
		linphone_core_set_capture_device(this->sipCore, this->audioInDevice.c_str());
	} else {
		linphone_core_set_capture_device(this->sipCore, "OSS: /dev/dsp");
	}
	if(!this->audioOutDevice.empty()) {
		if(this->audioOutDevice == "alsa:default") {
			this->audioOutDevice = "ALSA: default device";
		} else if(this->audioOutDevice == "/dev/dsp") {
			this->audioOutDevice = "OSS: /dev/dsp";
		}
		linphone_core_set_playback_device(this->sipCore, this->audioOutDevice.c_str());
	} else {
		linphone_core_set_playback_device(this->sipCore, "OSS: /dev/dsp");
	}

	MSList *audioCodecs = NULL;
	for(list<string>::iterator it = this->codecs.begin(); it != this->codecs.end(); ++it) {
		if(!ms_filter_codec_supported(it->c_str())) {
			cerr << "Codec " << *it << " is not supported by mediastreamer2! (ignored)" << endl;
			continue;
		}
		PayloadType *pt = payload_type_new();
		pt->mime_type = ms_strdup(it->c_str());
		pt->clock_rate = 8000; // TODO: make it configurable
		pt->flags |= (1 << 4); // PAYLOAD_TYPE_ENABLED
		audioCodecs = ms_list_append(audioCodecs, (void*)pt);
	}
	if(audioCodecs) {
		linphone_core_set_audio_codecs(this->sipCore, audioCodecs);
		linphone_core_set_audio_jittcomp(this->sipCore, 60);
//		linphone_core_update_allocated_audio_bandwidth(this->sipCore);
	}
#endif	

	if(!this->stunserverIP.empty()) {
		char buf[1024];
		if(this->stunserverPort) {
			snprintf(buf, sizeof(buf), "%s:%d", this->stunserverIP.c_str(), this->stunserverPort);
		} else {
			snprintf(buf, sizeof(buf), "%s", this->stunserverIP.c_str());
		}
		linphone_core_set_stun_server(this->sipCore, buf);
	}
}

bool SipbackendLinphone::shutdown(){
	this->quit = true;
	join();

	if(this->sipCore) {
		linphone_core_destroy(this->sipCore);
	}
	if(this->sipAuth) {
		linphone_auth_info_destroy(this->sipAuth);
	}
	if(this->sipProxy) {
		linphone_proxy_config_destroy(this->sipProxy);
	}

	return true;
}

bool SipbackendLinphone::registerAccount() {
	char buf[1024];

	/* clear old proxy config */
	linphone_core_clear_proxy_config(this->sipCore);
	
	/* clear old user info */
	linphone_core_clear_all_auth_info(this->sipCore);
	
	this->sipProxy = linphone_proxy_config_new();
	if(!this->sipProxy) {
		return false;
	}

	linphone_proxy_config_edit(this->sipProxy);
	snprintf(buf, sizeof(buf), "sip:%s", this->account.registrar.c_str());
	linphone_proxy_config_set_server_addr(this->sipProxy, buf);
	snprintf(buf, sizeof(buf), "sip:%s@%s", this->account.user.c_str(), this->account.registrar.c_str());
	linphone_proxy_config_set_identity(this->sipProxy, buf);
	linphone_proxy_config_enable_register(this->sipProxy, true);
	linphone_core_add_proxy_config(this->sipCore, this->sipProxy);
	linphone_proxy_config_done(this->sipProxy);

	this->sipAuth = linphone_auth_info_new(
			this->account.user.c_str(),
			NULL,
			this->account.passwd.c_str(),
			NULL,
			this->account.realm.c_str());
	if(!this->sipAuth) {
		return false;
	}
	linphone_core_add_auth_info(this->sipCore, this->sipAuth);

	return true;
}

void SipbackendLinphone::reRegisterAccount() {
	if(!readConfig()) {
		onError.emit("Configuration error!", true, "", 0);
		return;
	}

	if(this->sipCore) {
		if(this->sipAuth) {
			linphone_core_abort_authentication(this->sipCore, this->sipAuth);
			linphone_auth_info_destroy(this->sipAuth);
			this->sipAuth = NULL;
		}
		registerAccount();
	}
}

bool SipbackendLinphone::call(const std::string &user, const std::string &domain) {
	if(this->accountRegistered) {
		char buf[1024];
		snprintf(buf, sizeof(buf), "sip:%s@%s", user.c_str(), (domain.empty() ? this->account.registrar.c_str() : domain.c_str()));
		linphone_core_invite(this->sipCore, buf);
		return true;
	}

	return false;
}

bool SipbackendLinphone::hangup(){
	if(!this->accountRegistered) {
		return false;
	}
	
	return (linphone_core_terminate_call(this->sipCore, NULL) != -1);
}

bool SipbackendLinphone::answer(){
	if(this->accountRegistered) {
		if(linphone_core_inc_invite_pending(this->sipCore)) {
			linphone_core_accept_call(this->sipCore, NULL);
			return true;
		}
	}

	return false;
}

bool SipbackendLinphone::isInitialized(){
	return this->linphoneInitialized;
}

bool SipbackendLinphone::inCall() {
	return this->inCallStatus;
}

bool SipbackendLinphone::inTrying(){
	if(this->accountRegistered) {
		return linphone_core_inc_invite_pending(this->sipCore);
	}

	return false;
}

void SipbackendLinphone::threadMain() {
	initializeLinphone();
	registerAccount();

	while(!this->quit) {
		linphone_core_iterate(this->sipCore);
		sleep(1);
	}
}

void SipbackendLinphone::handleState(gstate_t state, const char *message) {
	switch(state) {
		case GSTATE_POWER_OFF:
			this->accountRegistered = false;
			this->linphoneInitialized = false;
			break;
		case GSTATE_POWER_STARTUP:
			break;
		case GSTATE_POWER_ON:
			this->linphoneInitialized = true;
			break;
		case GSTATE_POWER_SHUTDOWN:
			break;
		case GSTATE_REG_NONE:
			this->accountRegistered = false;
			this->onRegister.emit(FAILED);
			break;
		case GSTATE_REG_OK:
			this->accountRegistered = true;
			this->onRegister.emit(OK);
			break;
		case GSTATE_REG_PENDING:
			this->onRegister.emit(TRYING);
			break;
		case GSTATE_REG_FAILED:
			this->accountRegistered = false;
			this->onRegister.emit(FAILED);
			sleep(3);
			if(this->accountRegistered) {
				break;
			}
			reRegisterAccount();
			break;
		case GSTATE_CALL_IDLE:
			this->inCallStatus = false;
			break;
		case GSTATE_CALL_OUT_INVITE: {
			MMSEvent *e = new MMSEvent("SIP.backend.calling");
			e->send();
			break;
		}
		case GSTATE_CALL_OUT_CONNECTED:
			this->inCallStatus = true;
			onEstablished.emit(NULL);
			break;
		case GSTATE_CALL_IN_INVITE:
			DEBUGMSG("SIPBACKEND", "call incoming");
			this->answer();
			break;
		case GSTATE_CALL_IN_CONNECTED:
			this->inCallStatus = true;
			onEstablished.emit(NULL);
			break;
		case GSTATE_CALL_END:
			DEBUGMSG("SIPBACKEND", "remote hangup - call finished.");
			this->inCallStatus = false;
			onDisconnect.emit(NULL);
			break;
		case GSTATE_CALL_ERROR:
			this->inCallStatus = false;
			break;
		default:
			break;
	}
}

#endif /* __USE_LINPHONE__ */
