/***************************************************************************
 * This file is part of the BF527 doorbell application.                    *
 * Doorbell is a VOIP-based outdoor doorbell that can be run on SIP-based  *
 * telephone systems (PBX).                                                *
 *                                                                         *
 * Copyright (C) 2009,2010 by BerLinux Solutions GmbH                      *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at your  *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.           *
 ***************************************************************************/

#ifdef __USE_MINISIP__

#ifdef __MTRACE2__
#include <mcheck.h>
#endif

extern "C" {
#include <stdlib.h>

#include <libmnetutil/init.h>
#include <libmnetutil/NetUtil.h>
#include <libmnetutil/UDPSocket.h>
#include <libmnetutil/NetworkException.h>
#include <libmcrypto/init.h>
#include <libmsip/SipCommandString.h>
#include <libmsip/SipTransport.h>

#include <libminisip/media/Media.h>
#include <libminisip/media/MediaCommandString.h>
#include <libminisip/media/codecs/Codec.h>
#include <libminisip/media/soundcard/SoundDriverRegistry.h>
#include <libminisip/media/soundcard/Resampler.h>

#include <mmsbase/mmsevent.h>
#include <mmstools/tools.h>

#include "sipbackend_minisip.h"

#undef __ENABLE_LOG__

using namespace std;

SipbackendMinisip::SipbackendMinisip() :
	Sipbackend(),
	currentCallID(""),
	callUser(""),
	inviteCallID(""),
	minisipInitialized(false),
	muted(false),
	eventSignup(NULL),
#ifdef __HAVE_MIXER__
	audioCtrl(NULL),
#endif
	sipConf(NULL),
	subMedia(NULL),
	sip(NULL),
	sipStack(NULL) {
//	this->tryingTimer = new MMSTimer(true);
//	this->tryingTimer->timeOut.connect(sigc::mem_fun(this, &SipbackendMinisip::onTryingTimeOut));
}

bool SipbackendMinisip::initialize() {
	if(Sipbackend::initialize()) {
		initializeMiniSip();
		registerAccounts();
		return true;
	}

	return false;
}

SipbackendMinisip::~SipbackendMinisip() {
	shutdown();
#ifdef __HAVE_MIXER__
	delete audioCtrl;
#endif
}


bool SipbackendMinisip::initializeMiniSip() {
	DEBUGMSG("SIPBACKEND", "Initializing MiniSip");

#ifdef __HAVE_MIXER__
	try {
		audioCtrl = new MMSAudioCtrl();
	} catch(MMSAudioCtrlError &e) {
		DEBUGMSG("SIPBACKEND", "Failed to create MMSAudioCtrl: " + e.getMessage());
		audioCtrl = NULL;
	}
#endif

	MRef<AudioCodecRegistry* > audioCodecRegistry   = NULL;
	MRef<SoundDriverRegistry* > soundDriverRegistry = NULL;
	try {
		soundDriverRegistry = SoundDriverRegistry::getInstance();
		audioCodecRegistry = AudioCodecRegistry::getInstance();

		ConfigRegistry::getInstance();
		MediaRegistry::getInstance();
		ResamplerRegistry::getInstance();
		PhoneBookIoRegistry::getInstance();

#ifdef MINISIP_PLUGINDIR
		MRef<MPluginManager *> pluginManager = MPluginManager::getInstance();
		pluginManager->loadFromDirectory(MINISIP_PLUGINDIR);
#else
	#warning Missing MINISIP_PLUGINDIR
#endif

		libmnetutilInit();
		libmcryptoInit();

		DEBUGMSG("SIPBACKEND", "Configuring MiniSip");
		this->sipConf = new SipSoftPhoneConfiguration();
		if(this->audioInDevice != "") {
			this->sipConf->soundDeviceIn  = this->audioInDevice;
		} else {
			this->sipConf->soundDeviceIn  = "/dev/dsp";
		}
		if(this->audioOutDevice != "") {
			this->sipConf->soundDeviceOut = this->audioOutDevice;
		} else {
			this->sipConf->soundDeviceOut = "/dev/dsp";
		}
		this->sipConf->soundIOmixerType = "spatial";

		list<string>::iterator it;
		for(it = this->codecs.begin(); it != this->codecs.end(); ++it) {
			this->sipConf->audioCodecs.push_back(*it);
		}

		//this->sipConf->audioCodecs.push_back("G.711");
		//this->sipConf->audioCodecs.push_back("GSM");
		//this->sipConf->audioCodecs.push_back("G.711a");

		if(this->interface != "")
			this->sipConf->networkInterfaceName = this->interface;

		if(!this->stunserverIP.empty()) {
			this->sipConf->useSTUN = true;
			this->sipConf->stunServerIpString = this->stunserverIP;
			this->sipConf->stunServerPort = this->stunserverPort;
		}

#ifdef MINISIP_RINGTONE_SUPPORT
		if(this->ringtone != "")
			this->sipConf->ringtone = this->ringtone;
#endif

		DEBUGMSG("SIPBACKEND", "Initializing MiniSip-NetUtil");
		if(!NetUtil::init()) {
			DEBUGMSG("SIPBACKEND", "Could not initialize MiniSip-NetUtil");
			throw MMSSipBackendError(0, "Could not initialize MiniSip-NetUtil");
		}

		/* only enable UDP transport */
		list< MRef<SipTransportConfig*> >::iterator j;
		for(j  = this->sipConf->sipStackConfig->transports.begin();
		    j != this->sipConf->sipStackConfig->transports.end(); j++ ) {
			if((*j)->getName() == "UDP") {
				(*j)->setEnabled(true);
				break;
			}
		}

#ifdef __ENABLE_LOG__
		mdbg.setEnabled(true);
		mdbg.setPrintStreamName(true);
		mout.setPrintStreamName(true);
		merr.setPrintStreamName(true);
		this->consoleDbg = MRef<ConsoleDebugger*>(new ConsoleDebugger(this->sipConf));
		MRef<Thread *> consoleDbgThread = this->consoleDbg->start();
#else
		mdbg.setEnabled(false);
		mout.setEnabled(false);
		merr.setEnabled(false);
#endif

		DEBUGMSG("SIPBACKEND", "Creating IP-Provider");
		MRef<IpProvider *> ipProvider = IpProvider::create(this->sipConf);
		while(ipProvider->getExternalIp() == "127.0.0.1") {
			onError.emit("No network!", true, "Retry in %d s", 3);
			sleep(3);
			ipProvider = IpProvider::create(this->sipConf);
		}

		if(!this->sipConf->sipStackConfig)
			DEBUGMSG("SIPBACKEND", "sipStackConfig is NULL");

		bool done;
		int ntries = 8;
		int port = this->sipConf->sipStackConfig->preferedLocalSipPort = 20000;
		MRef<UDPSocket*> udpSocket;
		do {
			done = true;
			try {
				udpSocket = new UDPSocket(port);
			} catch(NetworkException &) {
				this->sipConf->sipStackConfig->preferedLocalSipPort = port = 32768 + rand()%32000;
				done = false;

			}
			ntries--;
		} while(!done && ntries > 0);
		
		if(!udpSocket) {
			DEBUGMSG("SIPBACKEND", "Error creating UDP socket");
			throw MMSSipBackendError(0, "Error creating UDP socket");
		}

		this->sipConf->sipStackConfig->externalContactUdpPort 	= ipProvider->getExternalPort(udpSocket);
		this->sipConf->sipStackConfig->localIpString 			= ipProvider->getExternalIp();
		this->sipConf->sipStackConfig->externalContactIP 		= this->sipConf->sipStackConfig->localIpString;

		DEBUGMSG("SIPBACKEND", "Creating media subsystem");
		this->subMedia = new SubsystemMedia(this->sipConf, ipProvider, NULL);

		DEBUGMSG("SIPBACKEND", "Creating SIP stack");
		this->sip = new Sip(this->sipConf, this->subMedia);

		bool codecFound = false;
		if(audioCodecRegistry) {
			for(list<string>::iterator it = this->codecs.begin(); it != this->codecs.end(); ++it) {
				if(audioCodecRegistry->create(*it) != 0) {
					codecFound = true;
					break;
				}
			}
		}
		if(!codecFound) {
			onError.emit("Codec mismatch!", true, "", 0);
		}

#if 0 /* getDeviceNames is not implemented in minisip */
		bool inDeviceFound  = false;
		bool outDeviceFound = false;
		if(soundDriverRegistry) {
			std::vector< MRef < SoundDriver * > > drivers = soundDriverRegistry->getDrivers();
			vector<MRef<SoundDriver *> >::iterator it = drivers.begin();
			vector<MRef<SoundDriver *> >::iterator end = drivers.end();
			vector<SoundDeviceName> soundDevices;
			vector<SoundDeviceName>::iterator itDevices;
			vector<SoundDeviceName>::iterator endDevices;

			for(; it != end; ++it) {
				printf("\nin loop\n");
				soundDevices = (*it)->getDeviceNames();
				itDevices = soundDevices.begin();
				endDevices = soundDevices.end();
				for(;itDevices != endDevices; ++itDevices) {
					printf("\n\nsoundDevice: %s\n%s\n\n", (*itDevices).getName().c_str(), (*itDevices).getDescription().c_str());
				}
			}
		}

		if(!inDeviceFound) {
			onError.emit("Audio input device " + this->audioInDevice + "not found", true, "", 0);
		}
		if(!outDeviceFound) {
			onError.emit("Audio output device " + this->audioInDevice + "not found", true, "", 0);
		}
#endif

		this->sipConf->sipStack = this->sipStack = this->sip->getSipStack();
		this->sipStack->setCallback(this);

#ifndef MINISIP_RINGTONE_SUPPORT
		this->ringtoneThread = new RingtoneThread(this->ringtone, this->busytone, this->calltone);
		if(this->ringtoneThread)
			this->ringtoneThread->start();
#endif
	}
	catch(exception &e) {
		DEBUGMSG("SIPBACKEND", string("Error initializing MiniSip: ") + e.what());
		throw MMSSipBackendError(0, string("Error initializing MiniSip: ") + e.what());
	}

#ifdef __ENABLE_LOG__
		this->consoleDbg->showConfig();
#endif

	this->minisipInitialized = true;

	return true;
}

bool SipbackendMinisip::shutdown() {
	DEBUGMSG("SIPBACKEND", "shutting down");

#ifdef __ENABLE_LOG__
	if(!this->consoleDbg.isNull() ) {
		this->consoleDbg->stop();
		this->consoleDbg->join();
		this->consoleDbg->setMediaHandler(NULL);
		this->consoleDbg = NULL;
	}
#endif

	if(!this->sip.isNull()) {
		CommandString cmdstr("", SipCommandString::sip_stack_shutdown);
		SipSMCommand sipcmd(cmdstr, SipSMCommand::dialog_layer, SipSMCommand::dispatcher);
		this->sipStack->handleCommand(sipcmd);
		sleep(1);
		this->sip->stop();
		this->sip = NULL;
	}
	
	return true;
}

bool SipbackendMinisip::registerAccounts() {
	SipUri userUri("sip:" + account.user + "@" + account.registrar);
	MRef<SipIdentity*> ident= new SipIdentity(userUri);
	ident->setIdentityName(account.user);

	MRef<SipCredential*> cred = new SipCredential(account.user, account.passwd, account.realm);
	ident->setCredential(cred);

	MRef<SipRegistrar*> registrar = new SipRegistrar(SipUri("sip:" + account.registrar), "UDP");
	ident->setSipRegistrar(registrar);
	ident->setDoRegister(true);
	string err = ident->setSipProxy(true, userUri.getString(), "UDP", "", 5060);
	if(!err.empty()) {
		this->onError.emit("Error detecting proxy: " + err, false, "", 0);
		DEBUGMSG("SIPBACKEND", "error detecting proxy: " + err);
	}
	
	this->sipConf->defaultIdentity = ident;
	this->sipConf->identities.push_back(ident);

	DEBUGMSG("SIPBACKEND", "Starting SIP thread");
	this->sip->start();

#ifdef __ENABLE_LOG__
	this->sipConf->sipStack->setDebugPrintPackets(true);
#else
	this->sipConf->sipStack->setDebugPrintPackets(false);
#endif

	return true;
}

void SipbackendMinisip::reRegisterAccount() {
	if(!readConfig()) {
		onError.emit("Configuration error!", true, "", 0);
		return;
	}
	if(!this->sip.isNull()) {
		MRef<SipStack*> stack = this->sip->getSipStack();
		if(!stack.isNull()) {
			CommandString reg("",SipCommandString::proxy_register);
			reg["proxy_domain"] = account.registrar;
			reg["identityId"] = this->sipConf->defaultIdentity->getId();
			MRef<SipCredential*> cred = this->sipConf->defaultIdentity->getCredential();
			cred->set(account.user, account.passwd, account.realm);
			this->sipConf->defaultIdentity->setCredential(cred);
			SipSMCommand sipcmd(reg, SipSMCommand::dialog_layer, SipSMCommand::dialog_layer);
			stack->handleCommand(sipcmd);
		}
	}
}

bool SipbackendMinisip::call(const string &user, const string &domain) {
#ifdef __MTRACE2__
	mtrace();
#endif
	if(!this->currentCallID.empty()) {
		printf("SIP.backend.call.error: Already call in progress\n");
		return false;
	}

	string uri = "sip:" + user + "@" + (domain != "" ? domain : this->sipConf->defaultIdentity->getSipUri().getIp());
	CommandString inv("", SipCommandString::invite, uri);
	CommandString resp = this->sipStack->handleCommandResp("sip", inv);
	this->inviteCallID = resp.getDestinationId();
	DEBUGMSG("SIPBACKEND", "Try to call, using inviteID " + this->inviteCallID);

	if(this->inviteCallID == "malformed") {
		this->inviteCallID = "";
		this->onError.emit(uri + " not found!", false, "", 0);
		DEBUGMSG("SIPBACKEND", "Invalid call target: " + uri);
		return false;
	}

//	this->tryingTimer->start(20000);

	return true;
}

bool SipbackendMinisip::hangup() {
	if(!accountRegistered) {
		return false;
	}

	setVolume(0);

	DEBUGMSG("SIPBACKEND", "hangup call");
	/* reject invited call */
	if(!this->inviteCallID.empty()) {
		DEBUGMSG("SIPBACKEND", "cancel");
		CommandString hup(this->inviteCallID, SipCommandString::cancel);
		this->sipStack->handleCommand(hup);
		this->inviteCallID = "";
		this->currentCallID = "";
		this->stopTone();
	} else if(!this->currentCallID.empty()) {
		/* hang up active call */
		DEBUGMSG("SIPBACKEND", "hangup");
		CommandString hup(this->currentCallID, SipCommandString::hang_up);
		this->sipStack->handleCommand(hup);
		this->inviteCallID = "";
		this->currentCallID = "";
	}

	return true;
}

/**
 * Answering a call with given id.
 *
 * @note	If another call is in progress, hangup()
 * will be called on it.
 *
 * @param 	id	[in]	call id to answer
 * @return	true, if answering was successful
 */
bool SipbackendMinisip::answer() {
	if(this->inviteCallID.empty()) {
		DEBUGMSG("SIPBACKEND", "SIP.backend.answer.error");
		DEBUGMSG("SIPBACKEND", "direction: incoming");
		DEBUGMSG("SIPBACKEND", "reason: No incoming call");
		return false;
	}

	if(!this->accountRegistered) {
		DEBUGMSG("SIPBACKEND", "SIP.backend.answer.error");
		DEBUGMSG("SIPBACKEND", "direction: incoming");
		DEBUGMSG("SIPBACKEND", "reason: Account not registered");
		return false;
	}

	/* hangup active call */
	if(!this->currentCallID.empty()) {
		CommandString hup(this->currentCallID, SipCommandString::hang_up);
		this->sipStack->handleCommand(hup);
		this->currentCallID = "";
	}

	CommandString answer(this->inviteCallID, SipCommandString::accept_invite);
	this->sipStack->handleCommand(answer);

	/* unmute senders */
	CommandString cmdstr(this->inviteCallID, MediaCommandString::set_session_sound_settings, "senders", "ON");
	this->subMedia->handleCommand("media", cmdstr);

	/* test mute receivers */
	this->subMedia->handleCommand("media", CommandString(this->inviteCallID, MediaCommandString::set_session_sound_settings, "receivers", "OFF"));

	return true;
}

void SipbackendMinisip::handleCommand(string subsystem, const CommandString &cmd) {
	const string op = cmd.getOp();

	cout << "SipbackendMinisip::handleCommand " << op << endl;
	DEBUGMSG("SIPBACKEND", "handleCommand: %s", op.c_str());

	/* ************************************ */
	/* successfully registered              */
	/* ************************************ */
	if(op == SipCommandString::register_ok) {
		this->accountRegistered = true;
		this->onRegister.emit(OK);
		DEBUGMSG("SIPBACKEND", "register ok for account %s", account.user.c_str());
	}
	/* ************************************************ */
	/* registering failed                               */
	/* ************************************************ */
	else if((op == SipCommandString::register_failed) ||
			(op == SipCommandString::register_failed_authentication)) {
		this->accountRegistered = false;
		DEBUGMSG("SIPBACKEND", "emitting error");
		this->onRegister.emit(FAILED);
		DEBUGMSG("SIPBACKEND", "registering failed for account %s", cmd.get("identityId").c_str());
		reRegisterAccount();
	}
	
	/* ************************************ */
	/* incoming call                        */
	/* ************************************ */
	else if(op == SipCommandString::incoming_available) {
		if (!inCall() && !inTrying()) {
			DEBUGMSG("SIPBACKEND", "call incoming %s", this->inviteCallID.c_str());
			this->inviteCallID = cmd.getDestinationId();
			this->callUser = cmd.getParam();
			this->answer();
		} else {
			CommandString hup(cmd.getDestinationId(), SipCommandString::reject_invite);
			this->sipStack->handleCommand(hup);
		}
	}
	/* ************************************ */
	/* remote ringing                       */
	/* ************************************ */
	else if(op == SipCommandString::remote_ringing) {
		MMSEvent *e = new MMSEvent("SIP.backend.calling");
		e->send();
		startCalltone();
	}
	/* ************************************ */
	/* call accepted                        */
	/* ************************************ */
	else if(op == SipCommandString::invite_ok) {
		this->stopTone();
//		this->tryingTimer->stop();
		/* unmute senders */
		CommandString cmdstr(this->inviteCallID, MediaCommandString::set_session_sound_settings, "senders", "ON");
		this->subMedia->handleCommand("media", cmdstr);

		DEBUGMSG("SIPBACKEND", "call accepted invite (%s)", cmd.getDestinationId().c_str());
		onEstablished.emit(NULL);
		this->currentCallID = cmd.getDestinationId();
		this->inviteCallID = "";
	}
	/* ************************************ */
	/* remote hangup                        */
	/* ************************************ */
	else if((op == SipCommandString::remote_hang_up) ||
			(op == SipCommandString::remote_cancelled_invite)) {
		this->stopTone();
		this->currentCallID = "";
#ifdef __MTRACE2__
		muntrace();
#endif
		DEBUGMSG("SIPBACKEND", "remote hangup - call finished.");
		onDisconnect.emit(NULL);
	}
	/* ************************************ */
	/* call terminated                      */
	/* ************************************ */
	else if(op == SipCommandString::call_terminated) {
		this->currentCallID = "";
		DEBUGMSG("SIPBACKEND", "call terminated.");
		onDisconnect.emit(NULL);
	}
	/* ************************************ */
	/* user not found                       */
	/* ************************************ */
	else if(op == SipCommandString::remote_user_not_found) {
		this->inviteCallID = "";
		this->onError.emit("User not found!", false, "", 0);
		DEBUGMSG("SIPBACKEND", "remote reject - user not found.");
	}
	/* ************************************ */
	/* remote reject                        */
	/* ************************************ */
	else if(op == SipCommandString::remote_reject) {
		this->stopTone();
		this->inviteCallID = "";
		onDisconnect.emit(NULL);
		DEBUGMSG("SIPBACKEND", "remote reject - call finished.");
	}
	/* ************************************ */
	/* transport error                      */
	/* ************************************ */
	else if(op == SipCommandString::transport_error) {
		if(!this->accountRegistered) {
			this->onRegister.emit(FAILED);
			reRegisterAccount();
		} else {
			this->onError.emit("RTP error!", false, "", 0);
		}
		DEBUGMSG("SIPBACKEND", "transport error");
	}
}

CommandString SipbackendMinisip::handleCommandResp(string subsystem, const CommandString &cmd) {
	CommandString ret("","");
	return ret;
}

#ifdef MINISIP_RINGTONE_SUPPORT
void SipbackendMinisip::startRingtone() {
	CommandString cmdstr = CommandString("", MediaCommandString::start_ringing);
	this->subMedia->handleCommand("media", cmdstr);
}

void SipbackendMinisip::startCalltone() {

}

void SipbackendMinisip::stopTone() {
	CommandString cmdstr = CommandString("", MediaCommandString::stop_ringing);
	this->subMedia->handleCommand("media", cmdstr);
}
#else
void SipbackendMinisip::startRingtone() {
	if(this->muted) {
	}	
	this->ringtoneThread->startRingtone();
}

void SipbackendMinisip::startCalltone() {
	this->ringtoneThread->startCalltone();
}

void SipbackendMinisip::stopTone() {
	if(this->muted) {
	}	
	this->ringtoneThread->stopTone();
}
#endif

bool SipbackendMinisip::isInitialized() {
	return this->accountRegistered;
}

bool SipbackendMinisip::inCall() {
	return !this->currentCallID.empty();
}

bool SipbackendMinisip::inTrying() {
	return !this->inviteCallID.empty();
}

void SipbackendMinisip::onTryingTimeOut() {
	if(this->currentCallID.empty()) {
		CommandString hup(this->inviteCallID, SipCommandString::cancel);
		this->sipStack->handleCommand(hup);
		this->inviteCallID = "";
		this->currentCallID = "";
		this->stopTone();
	}
}

void handleDTMF(unsigned int tone, unsigned int seq) {
	static unsigned int lastSeq = 0;

	if((seq != lastSeq) && (tone == 11)) {
		if(!system("scripts/blink.sh")) {
			DEBUGMSG("sipbackend", "error opening door");
		}
		lastSeq = seq;
	}
}

#endif /* __USE_MINISIP__ */
