/***************************************************************************
 * This file is part of the BF527 doorbell application.                    *
 * Doorbell is a VOIP-based outdoor doorbell that can be run on SIP-based  *
 * telephone systems (PBX).                                                *
 *                                                                         *
 * Copyright (C) 2009,2010 by BerLinux Solutions GmbH                      *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published by the   *
 * Free Software Foundation; either version 2 of the License, or (at your  *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it will be useful, but     *
 * WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       *
 * General Public License for more details.                                *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.           *
 ***************************************************************************/

#include <iostream>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cstring>
#include "sunphase.h"

static double pi2=6.283185307179586476925286766559;
static double pi=3.1415926535897932384626433832795;
static double RAD = 0.017453292519943295769236907684886;

static double julianDate(int year, int mon, int day,int h, int min, double s)
{
	int   greg;

	if(mon<=2){
		mon=mon +12;
		year = year -1;
	}

	greg = (year/400)-(year/100)+(year/4);
	return 2400000.5+365.0*year - 679004.0 + greg
		  + int(30.6001*(mon+1)) + day + h/24.0
		  + min/1440.0 + s/86400.0;
}


static double InPi(double x)
{
	int n = (int)(x/pi2);
	x = x - n*pi2;
	if (x<0) x +=pi2;
	return x;
}

static double eps(double T) {
	return RAD*(23.43929111 + (-46.8150*T - 0.00059*T*T + 0.001813*T*T*T)/3600.0);
}

static double computeTime(double &DK,double T) {

	double dRA, ramed, M, L, e, RA;

	ramed = 18.71506921 + 2400.0513369*T +(2.5862e-5 - 1.72e-9*T)*T*T;
	M = InPi(pi2 * (0.993133 + 99.997361*T));
	L = InPi(pi2 * (  0.7859453 + M/pi2 + (6893.0*sin(M)+72.0*sin(2.0*M)+6191.2*T) / 1296.0e3));
	e = eps(T);
	RA = atan(tan(L)*cos(e));


	if (RA<0.0) {
		RA+=pi;
	}
	if (L>pi) {
		RA+=pi;
	}

	RA = 24.0*RA/pi2;
	DK = asin(sin(e)*sin(L));
	ramed = 24.0*InPi(pi2*ramed/24.0)/pi2;

	dRA = ramed - RA;
	if(dRA < -12.0) {
		dRA+=24.0;
	}
	if(dRA > 12.0) {
		dRA-=24.0;
	}

	dRA = dRA * 1.0027379;

	return dRA ;
}

void calcSunPhase(time_t &rise, time_t &set, time_t now, double longitude, double latitude)
{
	double jd2000 = 2451545.0;
	double h = -50.0/60.0*RAD;
	struct tm struct_now;
	double jd, t, dk, timez, siderealday, difference,
	       riselocal, setlocal, riseutc, setutc, sunrise, sunset;
	int sunrisem,sunriseh,sunsetm,sunseth;

	latitude = latitude*RAD;

	/* get local time */
	memset(&struct_now,0,sizeof(struct_now));
	localtime_r(&now, &struct_now);


	timez = (double)(struct_now.tm_gmtoff / 3600);
	jd = julianDate(struct_now.tm_year + 1900,struct_now.tm_mon+1,struct_now.tm_mday,struct_now.tm_hour,struct_now.tm_min,0);
	t = (jd - jd2000)/36525.0;

	/* calc utc sunrise / set */
	siderealday = computeTime(dk,t);
	difference = 12.0*acos((sin(h) - sin(latitude)*sin(dk)) / (cos(latitude)*cos(dk)))/pi;
	riselocal = 12.0 - difference - siderealday;
	setlocal = 12.0 + difference - siderealday;
	riseutc = riselocal - longitude /15.0;
	setutc = setlocal - longitude /15.0;


	/* calc local sunrise */
	sunrise = riseutc + timez;
	if (sunrise<0.0) {
		sunrise +=24.0;
	} else if (sunrise>=24.0) {
		sunrise -=24.0;
	}

	/* calc local sunset */
	sunset = setutc + timez;
	if (sunset<0.0) {
		sunset +=24.0;
	} else if (sunset>=24.0) {
		sunset -=24.0;
	}

	/* calc regular hours and minutes */
	sunrisem = int(60.0*(sunrise - (int)sunrise)+0.5);
	sunriseh = (int)sunrise;
	if (sunrisem>=60){
		sunrisem-=60;
		sunriseh++;
	} else if (sunrisem<0) {
		sunrisem+=60;

		sunriseh--;
		if (sunriseh<0) {
			sunriseh+=24;
		}
	}

	sunsetm = int(60.0*(sunset - (int)sunset)+0.5);
	sunseth = (int)sunset;
	if (sunsetm>=60) {
		sunsetm-=60;
		sunseth++;
	} else if (sunsetm<0) {
		sunsetm+=60;
		sunseth--;

		if (sunsetm<0) {
			sunsetm+=24;
		}
	}

	struct_now.tm_hour=sunriseh;
	struct_now.tm_min=sunrisem;
	rise = mktime(&struct_now);
	struct_now.tm_hour=sunseth;
	struct_now.tm_min=sunsetm;
	set =  mktime(&struct_now);
}
